/*
 * Copyright (C) 2002 Robert Ernst <robert.ernst@linux-solutions.at>
 *
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See COPYING for GPL licensing information.
 *
 */

#include <qpe/qpeapplication.h>
#include <qpe/config.h>
#include <qtimer.h>
#include <qmessagebox.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qkeycode.h>
#include "GameWidget.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "images/data.c"

#define INTRO_COUNT	200	// screen-switch time in 50 ms cycles for intro
#define MOUSE_COUNT	6	// mouse autofire time in 50 ms cycles

GameWidget::GameWidget(QWidget *parent, const char *name, WFlags f) :
    QWidget(parent, name, f)
{
    /* set internal variables */
    m_lefthand = false;
    m_mouse_down = false;
    m_mouse_counter = 0;
    m_max_y = 0;
    m_max_x = 0;
    m_state = Intro;
    m_counter = 0;
    m_beam_counter = 0;
    m_bomb_counter = 0;
    memset(m_star, 0, sizeof (m_star));
    memset(m_bullet, 0, sizeof (m_bullet));
    memset(m_mine, 0, sizeof (m_mine));
    memset(m_enemy, 0, sizeof (m_enemy));
    memset(m_explosion, 0, sizeof (m_explosion));
    memset(m_bonus, 0, sizeof (m_bonus));
    memset(m_coord, 0, sizeof (m_coord));
    m_req_fire = 0;
    m_bonuses = 0;
    m_explosions = 0;
    m_bullets = 0;
    m_mines = 0;
    m_enemies = 0;
    m_gun = max_gun;
    m_shield = max_shield;
    m_fuel = max_fuel;
    m_bombs = 1;
    m_lives = 3;
    m_score = 0;
    m_enemies_killed = 0;
    m_motherships_killed = 0;
    m_bullets_fired = 0;
    m_bombs_fired = 0;
    memset(m_beam, 0, sizeof (m_beam));
    m_far_y = 0;
    m_near_y = 0;
    m_player_x = 0;
    m_player_req_x = 0;
    m_player_req_beam = 0;
    m_mothership_x = -100;
    m_mothership_req_x = -1;

    /* create the timer and pixmap resources */
    m_timer = new QTimer(this);
    connect(m_timer, SIGNAL(timeout()), this, SLOT(timerTick()));
    m_pixmap = 0;
    m_text[0].loadFromData(text_r_data_, sizeof (text_r_data_));
    m_text[1].loadFromData(text_l_data_, sizeof (text_l_data_));
    m_logo[0].loadFromData(logo_r_data_, sizeof (logo_r_data_));
    m_logo[1].loadFromData(logo_l_data_, sizeof (logo_l_data_));
    m_bomb_pm[0].loadFromData(bomb_0_data_, sizeof (bomb_0_data_));
    m_bomb_pm[1].loadFromData(bomb_1_data_, sizeof (bomb_1_data_));
    m_bonus_pm[0][0].loadFromData(bonus_0b_data_, sizeof (bonus_0b_data_));
    m_bonus_pm[1][0].loadFromData(bonus_1b_data_, sizeof (bonus_1b_data_));
    m_bonus_pm[2][0].loadFromData(bonus_2b_data_, sizeof (bonus_2b_data_));
    m_bonus_pm[3][0].loadFromData(bonus_3b_data_, sizeof (bonus_3b_data_));
    m_bonus_pm[0][1].loadFromData(bonus_0r_data_, sizeof (bonus_0r_data_));
    m_bonus_pm[1][1].loadFromData(bonus_1r_data_, sizeof (bonus_1r_data_));
    m_bonus_pm[2][1].loadFromData(bonus_2r_data_, sizeof (bonus_2r_data_));
    m_bonus_pm[3][1].loadFromData(bonus_3r_data_, sizeof (bonus_3r_data_));
    m_bonus_pm[0][2].loadFromData(bonus_0y_data_, sizeof (bonus_0y_data_));
    m_bonus_pm[1][2].loadFromData(bonus_1y_data_, sizeof (bonus_1y_data_));
    m_bonus_pm[2][2].loadFromData(bonus_2y_data_, sizeof (bonus_2y_data_));
    m_bonus_pm[3][2].loadFromData(bonus_3y_data_, sizeof (bonus_3y_data_));
    m_bonus_pm[0][3].loadFromData(bonus_0g_data_, sizeof (bonus_0g_data_));
    m_bonus_pm[1][3].loadFromData(bonus_1g_data_, sizeof (bonus_1g_data_));
    m_bonus_pm[2][3].loadFromData(bonus_2g_data_, sizeof (bonus_2g_data_));
    m_bonus_pm[3][3].loadFromData(bonus_3g_data_, sizeof (bonus_3g_data_));
    m_bonus_pm[0][4].loadFromData(bonus_0m_data_, sizeof (bonus_0m_data_));
    m_bonus_pm[1][4].loadFromData(bonus_1m_data_, sizeof (bonus_1m_data_));
    m_bonus_pm[2][4].loadFromData(bonus_2m_data_, sizeof (bonus_2m_data_));
    m_bonus_pm[3][4].loadFromData(bonus_3m_data_, sizeof (bonus_3m_data_));
    m_bullet_pm[0].loadFromData(bullet_0_data_, sizeof (bullet_0_data_));
    m_bullet_pm[1].loadFromData(bullet_1_data_, sizeof (bullet_1_data_));
    m_bullet_pm[2].loadFromData(bullet_2_data_, sizeof (bullet_2_data_));
    m_bullet_pm[3].loadFromData(bullet_3_data_, sizeof (bullet_3_data_));
    m_mine_pm[0].loadFromData(mine_0_data_, sizeof (mine_0_data_));
    m_mine_pm[1].loadFromData(mine_1_data_, sizeof (mine_1_data_));
    m_mine_pm[2].loadFromData(mine_2_data_, sizeof (mine_2_data_));
    m_mine_pm[3].loadFromData(mine_3_data_, sizeof (mine_3_data_));
    m_glider_pm[0].loadFromData(glider_r_data_, sizeof (glider_r_data_));
    m_glider_pm[1].loadFromData(glider_l_data_, sizeof (glider_l_data_));
    m_mothership_pm.loadFromData(mothership_data_, sizeof (mothership_data_));
    m_enemy_pm[0].loadFromData(enemy_0_data_, sizeof (enemy_0_data_));
    m_enemy_pm[1].loadFromData(enemy_1_data_, sizeof (enemy_1_data_));
    m_enemy_pm[2].loadFromData(enemy_2_data_, sizeof (enemy_2_data_));
    m_enemy_pm[3].loadFromData(enemy_3_data_, sizeof (enemy_3_data_));
    m_enemy_pm[4].loadFromData(enemy_4_data_, sizeof (enemy_4_data_));
    m_enemy_pm[5].loadFromData(enemy_5_data_, sizeof (enemy_5_data_));
    m_big_expl_pm[0].loadFromData(big_expl_0_data_, sizeof (big_expl_0_data_));
    m_big_expl_pm[1].loadFromData(big_expl_1_data_, sizeof (big_expl_1_data_));
    m_big_expl_pm[2].loadFromData(big_expl_2_data_, sizeof (big_expl_2_data_));
    m_big_expl_pm[3].loadFromData(big_expl_3_data_, sizeof (big_expl_3_data_));
    m_big_expl_pm[4].loadFromData(big_expl_4_data_, sizeof (big_expl_4_data_));
    m_big_expl_pm[5].loadFromData(big_expl_5_data_, sizeof (big_expl_5_data_));
    m_big_expl_pm[6].loadFromData(big_expl_6_data_, sizeof (big_expl_6_data_));
    m_big_expl_pm[7].loadFromData(big_expl_7_data_, sizeof (big_expl_7_data_));
    m_big_expl_pm[8].loadFromData(big_expl_8_data_, sizeof (big_expl_8_data_));
    m_big_expl_pm[9].loadFromData(big_expl_9_data_, sizeof (big_expl_9_data_));
    m_big_expl_pm[10].loadFromData(big_expl_10_data_, sizeof (big_expl_10_data_));
    m_big_expl_pm[11].loadFromData(big_expl_11_data_, sizeof (big_expl_11_data_));
    m_big_expl_pm[12].loadFromData(big_expl_12_data_, sizeof (big_expl_12_data_));
    m_big_expl_pm[13].loadFromData(big_expl_13_data_, sizeof (big_expl_13_data_));
    m_big_expl_pm[14].loadFromData(big_expl_14_data_, sizeof (big_expl_14_data_));
    m_big_expl_pm[15].loadFromData(big_expl_15_data_, sizeof (big_expl_15_data_));
    m_big_expl_pm[16].loadFromData(big_expl_16_data_, sizeof (big_expl_16_data_));
    m_big_expl_pm[17].loadFromData(big_expl_17_data_, sizeof (big_expl_17_data_));
    m_big_expl_pm[18].loadFromData(big_expl_18_data_, sizeof (big_expl_18_data_));
    m_big_expl_pm[19].loadFromData(big_expl_19_data_, sizeof (big_expl_19_data_));
    m_big_expl_pm[20].loadFromData(big_expl_20_data_, sizeof (big_expl_20_data_));
    m_big_expl_pm[21].loadFromData(big_expl_21_data_, sizeof (big_expl_21_data_));
    m_big_expl_pm[22].loadFromData(big_expl_22_data_, sizeof (big_expl_22_data_));
    m_small_expl_pm[0].loadFromData(small_expl_0_data_, sizeof (small_expl_0_data_));
    m_small_expl_pm[1].loadFromData(small_expl_1_data_, sizeof (small_expl_1_data_));
    m_small_expl_pm[2].loadFromData(small_expl_2_data_, sizeof (small_expl_2_data_));
    m_small_expl_pm[3].loadFromData(small_expl_3_data_, sizeof (small_expl_3_data_));
    m_small_expl_pm[4].loadFromData(small_expl_4_data_, sizeof (small_expl_4_data_));
    m_small_expl_pm[5].loadFromData(small_expl_5_data_, sizeof (small_expl_5_data_));
    m_small_expl_pm[6].loadFromData(small_expl_6_data_, sizeof (small_expl_6_data_));
    m_small_expl_pm[7].loadFromData(small_expl_7_data_, sizeof (small_expl_7_data_));
    m_small_expl_pm[8].loadFromData(small_expl_8_data_, sizeof (small_expl_8_data_));
    m_small_expl_pm[9].loadFromData(small_expl_9_data_, sizeof (small_expl_9_data_));
    m_small_expl_pm[10].loadFromData(small_expl_10_data_, sizeof (small_expl_10_data_));
    m_small_expl_pm[11].loadFromData(small_expl_11_data_, sizeof (small_expl_11_data_));
    m_small_expl_pm[12].loadFromData(small_expl_12_data_, sizeof (small_expl_12_data_));
    m_small_expl_pm[13].loadFromData(small_expl_13_data_, sizeof (small_expl_13_data_));
    m_small_expl_pm[14].loadFromData(small_expl_14_data_, sizeof (small_expl_14_data_));
    m_small_expl_pm[15].loadFromData(small_expl_15_data_, sizeof (small_expl_15_data_));
    m_small_expl_pm[16].loadFromData(small_expl_16_data_, sizeof (small_expl_16_data_));
    m_small_expl_pm[17].loadFromData(small_expl_17_data_, sizeof (small_expl_17_data_));
    m_small_expl_pm[18].loadFromData(small_expl_18_data_, sizeof (small_expl_18_data_));
    m_small_expl_pm[19].loadFromData(small_expl_19_data_, sizeof (small_expl_19_data_));
    m_small_expl_pm[20].loadFromData(small_expl_20_data_, sizeof (small_expl_20_data_));
    m_small_expl_pm[21].loadFromData(small_expl_21_data_, sizeof (small_expl_21_data_));
    m_small_expl_pm[22].loadFromData(small_expl_22_data_, sizeof (small_expl_22_data_));

    /* read the configuration file */
    readConfig();
}

GameWidget::~GameWidget()
{
    /* write the configuration file */
    writeConfig();

    /* delete the timer and pixmap resources */
    delete m_timer;
    delete m_pixmap;
}

void GameWidget::timerTick(void)
{
    m_counter++;
    if (m_mouse_down) {
	m_mouse_counter++;
	if (m_mouse_counter > MOUSE_COUNT) {
	    fireBullet();
	    m_mouse_counter = 0;
	}
    }
    switch (m_state) {
    case Intro:
	break;
    case Playing:
	m_beam_counter++;
	updateActors();
	break;
    case Dying:
	m_beam_counter++;
	updateActors();
	if (m_lives) {
	    if (m_counter > 30) {
		switchState(SafeMode);
		m_shield = max_shield;
		m_gun = max_gun;
		m_fuel = max_fuel;
	    }
	} else {
	    if (m_counter > 60) {
		switchState(GameOver);
	    }
	}
	break;
    case SafeMode:
	m_beam_counter++;
	updateActors();
	if (m_counter > 30) {
	    switchState(Playing);
	}
	break;
    case GameOver:
	if (m_counter > 600) {
	    switchState(Intro);
	}
	break;
    }
    repaint(false);
}

void GameWidget::paintEvent(QPaintEvent *event)
{
    if (m_state == Intro) {
	drawIntro();
    } else if (m_state == GameOver) {
	drawGameOver();
    } else {
	drawPlaying();
    }
}

void GameWidget::showEvent(QShowEvent *event)
{
    QWidget::showEvent(event);
    m_timer->start(50);
    setFocusPolicy(StrongFocus);
}

void GameWidget::hideEvent(QHideEvent *event)
{
    QWidget::hideEvent(event);
    m_timer->stop();
}

void GameWidget::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
    if (!m_pixmap || event->size() != event->oldSize()) {
	m_max_y = event->size().width();
	m_max_x = event->size().height();
	if (m_pixmap) {
	    delete m_pixmap;
	}
	m_pixmap = new QPixmap(m_max_y, m_max_x);
	calculateBeams();
    }
}

void GameWidget::keyPressEvent(QKeyEvent *event)
{
    QWidget::keyPressEvent(event);
    switch (event->key()) {
    case Key_Up:
	if (m_state == Intro) {
	    if (m_lefthand) {
		m_counter = m_counter + INTRO_COUNT - (m_counter % INTRO_COUNT);
	    } else {
		m_counter = m_counter + 2 * INTRO_COUNT - (m_counter % INTRO_COUNT);
	    }
	} else {
	    if (m_lefthand) {
		moveRight();
	    } else {
		moveLeft();
	    }
	}
	break;
    case Key_Down:
	if (m_state == Intro) {
	    if (m_lefthand) {
		m_counter = m_counter + 2 * INTRO_COUNT - (m_counter % INTRO_COUNT);
	    } else {
		m_counter = m_counter + INTRO_COUNT - (m_counter % INTRO_COUNT);
	    }
	} else {
	    if (m_lefthand) {
		moveLeft();
	    } else {
		moveRight();
	    }
	}
	break;
    case Key_Left:
	if (m_state == Intro || m_state == GameOver) {
	    flipScreen();
	} else if (m_lefthand) {
	    fireBullet();
	} else {
	    dropBomb();
	}
	break;
    case Key_Right:
	if (m_state == Intro || m_state == GameOver) {
	    flipScreen();
	} else if (m_lefthand) {
	    dropBomb();
	} else {
	    fireBullet();
	}
	break;
    case Key_Space:
    case Key_Enter:
	if (m_state == Intro || m_state == GameOver) {
	    start();
	} else {
	    fireBullet();
	}
	break;
    case Key_Escape:
	stop();
	break;
    }
}

void GameWidget::mousePressEvent(QMouseEvent *event)
{
    QWidget::mousePressEvent(event);
    if (m_state == Intro || m_state == GameOver) {
	start();
    } else {
	int x;
	int y;
	if (m_lefthand) {
	    x = m_max_x - event->y();
	    y = event->x();
	} else {
	    x = event->y();
	    y = m_max_y - event->x();
	}
	if (x < m_player_x) {
	    moveLeft();
	} else if (x > m_player_x) {
	    moveRight();
	}
	if (y < (m_max_y / 2)) {
	    fireBullet();
	}
	m_mouse_down = true;
	m_mouse_counter = 0;
    }
}

void GameWidget::mouseReleaseEvent(QMouseEvent *event)
{
    m_mouse_down = false;
    m_mouse_counter = 0;
}

void GameWidget::drawPixmapText(QPainter &painter, int x, int y, int wrap, const char *str)
{
    int xstart = x;

    while (*str != '\0') {
	if ((*str >= '!') && (*str <= 'Z')) {
	    if (m_lefthand) {
		painter.drawPixmap(y, m_max_x - x - 6, m_text[1], 0, 348 - 6 - (*str - '!') * 6, 9, 6);
	    } else {
		painter.drawPixmap(m_max_y - y - 9, x, m_text[0], 0, (*str - '!') * 6, 9, 6);
	    }
	    x = x + 7;
	} else if (*str == ' ') {
	    x = x + 7;
	    if (wrap > xstart) {
		int nchars = 0;
		const char *ptr;
		for (ptr = str + 1; *ptr >= '!' && *ptr <= 'Z'; ptr++) {
		    nchars++;
		}
		if (x + (nchars * 7) > wrap) {
		    x = xstart;
		    y = y + 10;
		}
	    }
	} else {
	    x = xstart;
	    y = y + 10;
	}
	str++;
    }
}

void GameWidget::drawPixmapTextCentered(QPainter &painter, int x, int y, int wrap, const char *str)
{
    int textwidth = 6 + (strlen(str) - 1) * 7;

    drawPixmapText(painter, x + (wrap - x - textwidth) / 2, y, 0, str);
}

void GameWidget::drawPixmapTextRightJustified(QPainter &painter, int x, int y, int wrap, const char *str)
{
    int textwidth = 6 + (strlen(str) - 1) * 7;

    drawPixmapText(painter, x + wrap - textwidth, y, 0, str);
}

void GameWidget::drawBar(QPainter &painter, int x, int y, int width, int value,
    int red, int yellow, int max)
{
    int divide = (value * width) / max;

    if (value <= red) {
	painter.setPen(QColor(255, 0, 0));
    } else if (value <= yellow) {
	painter.setPen(QColor(255, 255, 0));
    } else {
	painter.setPen(QColor(0, 255, 0));
    }
    if (divide < width) {
	if (m_lefthand) {
	    painter.drawRect(y, m_max_x - x - width, 8, width);
	} else {
	    painter.drawRect(m_max_y - y - 8, x, 8, width);
	}
    }
    if (divide > 0) {
	if (m_lefthand) {
	    painter.fillRect(y, m_max_x - x - divide, 8, divide, painter.pen().color());
	} else {
	    painter.fillRect(m_max_y - y - 8, x, 8, divide, painter.pen().color());
	}
    }
    painter.setPen(QColor(255, 255, 255));
}

void GameWidget::readConfig(void)
{
    Config cfg("Glider");

    cfg.setGroup("Settings");

    int lefthand = cfg.readNumEntry("Lefthand", 0);
    if (lefthand == 1) {
	m_lefthand = 1;
    } else {
	m_lefthand = 0;
    }
}

void GameWidget::writeConfig(void)
{
    Config cfg("Glider");

    cfg.setGroup("Settings");
    cfg.writeEntry("Lefthand", m_lefthand);
}

void GameWidget::moveLeft(void)
{
    int near_diff_x = m_beam[1].near_x - m_beam[0].near_x;

    m_req_fire = 0;
    if (m_fuel > 0 && (m_state == Playing || m_state == SafeMode)
	&& m_player_req_x >= m_player_x && m_player_x > m_beam[0].near_x) {
	m_player_req_x = m_player_req_x - near_diff_x;
	m_player_req_beam--;
	m_fuel = m_fuel - 1;
    }
}

void GameWidget::moveRight(void)
{
    int near_diff_x = m_beam[1].near_x - m_beam[0].near_x;

    m_req_fire = 0;
    if (m_fuel > 0 && (m_state == Playing || m_state == SafeMode)
	&& m_player_req_x <= m_player_x && m_player_x < m_beam[num_beams - 1].near_x) {
	m_player_req_x = m_player_req_x + near_diff_x;
	m_player_req_beam++;
	m_fuel = m_fuel - 1;
    }
}

void GameWidget::fireBullet(void)
{
    m_req_fire = 0;
    if (m_gun > 0 && m_bullets < max_bullets && (m_state == Playing || m_state == SafeMode)) {
	if (abs(m_beam[m_player_req_beam].near_x - m_player_x) < 4) {
	    m_bullet[m_bullets].beam = m_player_req_beam;
	    m_bullet[m_bullets].counter = 0;
	    m_bullets++;
	    m_gun = m_gun - 10;
	    m_bullets_fired++;
	} else if (m_player_req_beam > 0
	    && abs(m_beam[m_player_req_beam - 1].near_x - m_player_x) < 4) {
	    m_bullet[m_bullets].beam = m_player_req_beam - 1;
	    m_bullet[m_bullets].counter = 0;
	    m_bullets++;
	    m_gun = m_gun - 10;
	    m_bullets_fired++;
	} else if (m_player_req_beam < (num_beams - 1)
	    && abs(m_beam[m_player_req_beam + 1].near_x - m_player_x) < 4) {
	    m_bullet[m_bullets].beam = m_player_req_beam + 1;
	    m_bullet[m_bullets].counter = 0;
	    m_bullets++;
	    m_gun = m_gun - 10;
	    m_bullets_fired++;
	} else {
	    m_req_fire++;
	}
    }
}

void GameWidget::deleteBullet(int i)
{
    m_bullets--;
    while (i < m_bullets) {
	m_bullet[i] = m_bullet[i + 1];
	i++;
    }
}

void GameWidget::dropMine(int beam, int counter)
{
    if (m_mines < max_mines) {
	m_mine[m_mines].beam = beam;
	m_mine[m_mines].counter = counter;
	m_mines++;
    }
}

void GameWidget::deleteMine(int i)
{
    m_mines--;
    while (i < m_mines) {
	m_mine[i] = m_mine[i + 1];
	i++;
    }
}

void GameWidget::makeExplosion(int x, int y, int big)
{
    if (m_explosions < max_explosions) {
	m_explosion[m_explosions].x = x;
	m_explosion[m_explosions].y = y;
	m_explosion[m_explosions].big = big;
	m_explosion[m_explosions].counter = 0;
	m_explosions++;
    }
}

void GameWidget::deleteExplosion(int i)
{
    m_explosions--;
    while (i < m_explosions) {
	m_explosion[i] = m_explosion[i + 1];
	i++;
    }
}

void GameWidget::makeBonus(int beam, int counter, int always)
{
    if (m_bonuses < max_bonuses && (always || (rand() % 10) == 0)) {
	m_bonus[m_bonuses].beam = beam;
	m_bonus[m_bonuses].counter = counter;
	m_bonus[m_bonuses].what = rand() % 5;
	m_bonuses++;
    }
}

void GameWidget::deleteBonus(int i)
{
    m_bonuses--;
    while (i < m_bonuses) {
	m_bonus[i] = m_bonus[i + 1];
	i++;
    }
}

void GameWidget::makeEnemy(void)
{
    if (m_enemies < max_enemies) {
	if (rand() % 2) {
	    m_enemy[m_enemies].x = m_max_x + 7;
	    m_enemy[m_enemies].req_end_x = -7;
	} else {
	    m_enemy[m_enemies].x = -7;
	    m_enemy[m_enemies].req_end_x = m_max_x + 7;
	}
	m_enemy[m_enemies].beam = rand() % num_beams;
	m_enemy[m_enemies].req_x = m_beam[m_enemy[m_enemies].beam].far_x;
	m_enemy[m_enemies].counter = num_counter;
	m_enemy[m_enemies].req_counter = 75 - (rand() % 50);
	m_enemies++;
    }
}

void GameWidget::deleteEnemy(int i)
{
    m_enemies--;
    while (i < m_enemies) {
	m_enemy[i] = m_enemy[i + 1];
	i++;
    }
}

void GameWidget::dropBomb(void)
{
    if (m_bombs > 0 && m_bomb_counter == 0 && (m_state == Playing || m_state == SafeMode)) {
	m_bomb_counter = 20;
	m_bomb_x = m_player_x;
	m_bomb_y = m_near_y - 18;
	m_bomb_diff_x = ((m_max_x / 2) - m_player_x) / 20;
	m_bomb_diff_y = (m_near_y - 18 - m_far_y) / 60;
	m_bombs--;
	m_bombs_fired++;
    }
}

void GameWidget::flipScreen(void)
{
    m_lefthand = !m_lefthand;
    repaint(false);
}

void GameWidget::switchState(enum States state)
{
    if (m_state != state) {
	m_req_fire = 0;
	m_state = state;
	m_counter = 0;
    }
}

void GameWidget::start(void)
{
    switchState(Playing);
    m_beam_counter = 0;
    m_bomb_counter = 0;
    makeStars();
    memset(m_bullet, 0, sizeof (m_bullet));
    memset(m_mine, 0, sizeof (m_mine));
    memset(m_enemy, 0, sizeof (m_enemy));
    memset(m_bonus, 0, sizeof (m_bonus));
    memset(m_explosion, 0, sizeof (m_explosion));
    m_req_fire = 0;
    m_enemies = 0;
    m_explosions = 0;
    m_bullets = 0;
    m_mines = 0;
    m_gun = max_gun;
    m_shield = max_shield;
    m_fuel = max_fuel;
    m_bombs = 1;
    m_lives = 3;
    m_score = 0;
    m_enemies_killed = 0;
    m_motherships_killed = 0;
    m_bullets_fired = 0;
    m_bombs_fired = 0;
    m_player_x = m_beam[num_beams / 2].near_x;
    m_player_req_x = m_player_x;
    m_player_req_beam = num_beams / 2;
    m_mothership_x = -100;
    m_mothership_req_x = -1;
    repaint(false);
}

void GameWidget::stop(void)
{
    switchState(Intro);
    repaint(false);
    parentWidget()->close();
}

void GameWidget::drawIntro(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    int i;

    painter_pixmap.fillRect(0, 0, m_max_y, m_max_x, black);
    painter_pixmap.setPen(white);

    switch ((m_counter / INTRO_COUNT) % 3) {
    case 0:
	if (m_lefthand) {
	    painter_pixmap.drawPixmap(0, (m_max_x - m_logo[1].height()) / 2, m_logo[1]);
	} else {
	    painter_pixmap.drawPixmap(40, (m_max_x - m_logo[1].height()) / 2, m_logo[0]);
	}
	break;
    case 1:
	if (m_lefthand) {
	    painter_pixmap.drawPixmap(20, m_max_x - 20 - 35, m_glider_pm[1]);
	    painter_pixmap.drawPixmap(100, m_max_x - 20 - 29, m_mothership_pm);
	    painter_pixmap.drawPixmap(130, m_max_x - 20 - 15, m_enemy_pm[0]);
	    painter_pixmap.drawPixmap(160, m_max_x - 20 - 7, m_mine_pm[0]);
	    drawPixmapText(painter_pixmap, 65, 20, m_max_x - 20, tr("NAVIGATING\n\n"
		"JUMP BETWEEN THE BEAMS USING CURSOR UP/DOWN KEYS. "
		"USE SPACE OR CURSOR RIGHT TO FIRE A BULLET, CURSOR LEFT TO DROP A BOMB"));
	} else {
	    painter_pixmap.drawPixmap(m_max_y - 20 - 18, 20, m_glider_pm[0]);
	    painter_pixmap.drawPixmap(m_max_y - 100 - 15, 20, m_mothership_pm);
	    painter_pixmap.drawPixmap(m_max_y - 130 - 15, 20, m_enemy_pm[0]);
	    painter_pixmap.drawPixmap(m_max_y - 160 - 15, 20, m_mine_pm[0]);
	    drawPixmapText(painter_pixmap, 65, 20, m_max_x - 20, tr("NAVIGATING\n\n"
		"JUMP BETWEEN THE BEAMS USING  CURSOR UP/DOWN KEYS. "
		"USE SPACE OR CURSOR LEFT TO FIRE A BULLET, CURSOR RIGHT TO DROP A BOMB"));
	}
	drawPixmapText(painter_pixmap, 65, 100, m_max_x - 20, tr("THE ENEMIES\n\n"
	    "THE MINERS TRY TO KILL YOU BY DROPPING A MINE ON THE BEAM YOU ARE USING\n"
	    "THE MOTHERSHIP HAS NO WEAPONS, BUT YOU GET A BONUS IF YOU DESTROY IT"));
	break;
    case 2:
	for (i = 0; i < 5; i++) {
	    if (m_lefthand) {
		painter_pixmap.drawPixmap(20 + 35 * i, m_max_x - 20 - 23, m_bonus_pm[0][i]);
	    } else {
		painter_pixmap.drawPixmap(m_max_y - 20 - 35 * i - 11, 20, m_bonus_pm[0][i]);
	    }
	}
	drawPixmapText(painter_pixmap, 60, 20, m_max_x - 20,
	    tr("THIS BONUS INCREASES YOUR GUN STRENGTH"));
	drawPixmapText(painter_pixmap, 60, 55, m_max_x - 20,
	    tr("THIS BONUS INCREASES YOUR SHIELD STRENGTH"));
	drawPixmapText(painter_pixmap, 60, 90, m_max_x - 20,
	    tr("THIS BONUS INCREASES YOUR FUEL"));
	drawPixmapText(painter_pixmap, 60, 125, m_max_x - 20,
	    tr("THIS BONUS GIVES YOU A BOMB THAT DESTROYS ALL ENEMIES"));
	drawPixmapText(painter_pixmap, 60, 160, m_max_x - 20,
	    tr("THIS BONUS GIVES YOU AN EXTRA LIVE"));
	break;
    }
    drawPixmapTextCentered(painter_pixmap, 0, m_max_y - 36, m_max_x,
	tr("PRESS CURSOR UP/DOWN TO FLIP DISPLAY"));
    drawPixmapTextCentered(painter_pixmap, 0, m_max_y - 26, m_max_x,
	tr("PRESS SPACE TO START"));

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void GameWidget::drawGameOver(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);
    char buffer[80];

    painter_pixmap.fillRect(0, 0, m_max_y, m_max_x, black);
    painter_pixmap.setPen(white);

    if (m_lefthand) {
	painter_pixmap.drawPixmap(20, m_max_x - 20 - 35, m_glider_pm[1]);
	painter_pixmap.drawPixmap(80, m_max_x - 20 - 29, m_mothership_pm);
	painter_pixmap.drawPixmap(110, m_max_x - 20 - 15, m_enemy_pm[0]);
	painter_pixmap.drawPixmap(140, m_max_x - 20 - 7, m_mine_pm[0]);
    } else {
	painter_pixmap.drawPixmap(m_max_y - 20 - 18, 20, m_glider_pm[0]);
	painter_pixmap.drawPixmap(m_max_y - 80 - 15, 20, m_mothership_pm);
	painter_pixmap.drawPixmap(m_max_y - 110 - 15, 20, m_enemy_pm[0]);
	painter_pixmap.drawPixmap(m_max_y - 140 - 15, 20, m_mine_pm[0]);
    }

    snprintf(buffer, sizeof (buffer), tr("GAME OVER AFTER %d SECONDS!"), m_beam_counter / 20);
    drawPixmapTextCentered(painter_pixmap, 75, 20, m_max_x - 20, buffer);
    snprintf(buffer, sizeof (buffer), tr("YOU SCORED %d POINTS"), m_score);
    drawPixmapTextCentered(painter_pixmap, 75, 30, m_max_x - 20, buffer);
    snprintf(buffer, sizeof (buffer), tr("ENEMIES KILLED: %d"), m_enemies_killed);
    drawPixmapTextCentered(painter_pixmap, 75, 80, m_max_x - 20, buffer);
    snprintf(buffer, sizeof (buffer), tr("MOTHERSHIPS DESTROYED: %d"), m_motherships_killed);
    drawPixmapTextCentered(painter_pixmap, 75, 100, m_max_x - 20, buffer);
    snprintf(buffer, sizeof (buffer), tr("BULLETS FIRED: %d"), m_bullets_fired);
    drawPixmapTextCentered(painter_pixmap, 75, 120, m_max_x - 20, buffer);
    snprintf(buffer, sizeof (buffer), tr("BOMBS DROPPED: %d"), m_bombs_fired);
    drawPixmapTextCentered(painter_pixmap, 75, 140, m_max_x - 20, buffer);

    drawPixmapTextCentered(painter_pixmap, 0, m_max_y - 36, m_max_x,
	tr("PRESS CURSOR UP/DOWN TO FLIP DISPLAY"));
    drawPixmapTextCentered(painter_pixmap, 0, m_max_y - 26, m_max_x,
	tr("PRESS SPACE TO START"));

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void GameWidget::drawPlaying(void)
{
    QPainter painter_pixmap(m_pixmap);
    QPainter painter_widget(this);
    QColor black(0, 0, 0);
    QColor white(255, 255, 255);

    if (m_bomb_counter == 1) {
	painter_pixmap.fillRect(0, 0, m_max_y, m_max_x, white);
    } else {
	painter_pixmap.fillRect(0, 0, m_max_y, m_max_x, black);
    }
    painter_pixmap.setPen(white);

    drawStats(painter_pixmap);
    drawBeams(painter_pixmap);
    drawStars(painter_pixmap);
    drawMothership(painter_pixmap);
    drawBonuses(painter_pixmap);
    drawEnemies(painter_pixmap);
    drawBullets(painter_pixmap);
    drawMines(painter_pixmap);
    drawBomb(painter_pixmap);
    drawShip(painter_pixmap);
    drawExplosions(painter_pixmap);

    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, *m_pixmap);
}

void GameWidget::makeStars(void)
{
    int colors[] = {0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 3};
    int sizes[] = {1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4};
    int i;

    memset(m_star, 0, sizeof (m_star));
    for (i = 0; i < max_stars; i++) {
	m_star[i].x = rand() % m_max_x;
	m_star[i].y = rand() % (m_max_y - 20);
	m_star[i].size = sizes[rand() % 12];
	m_star[i].color = colors[rand() % 12];
    }
    if (m_lefthand) {
	for (i = 0; i < max_stars; i++) {
	    m_star[i].y = m_max_y - m_star[i].y;
	}
    }
}

void GameWidget::drawStars(QPainter &painter)
{
    int color;
    int size;
    int i;

    for (i = 0; i < max_stars; i++) {
	for (size = 0; size < m_star[i].size; size++) {
	    color = 255 / (size + 1);
	    switch (m_star[i].color) {
	    case 0:
		painter.setPen(QColor(color, color, color));
		break;
	    case 1:
		painter.setPen(QColor(color / 2, color / 2, color));
		break;
	    case 2:
		painter.setPen(QColor(color, color, 0));
		break;
	    case 3:
		painter.setPen(QColor(color, 0, 0));
		break;
	    }
	    if (size > 0) {
		painter.drawPoint(m_star[i].y, m_star[i].x + size);
		painter.drawPoint(m_star[i].y, m_star[i].x - size);
		painter.drawPoint(m_star[i].y + size, m_star[i].x);
		painter.drawPoint(m_star[i].y - size, m_star[i].x);
	    } else {
		painter.drawPoint(m_star[i].y, m_star[i].x);
	    }
	}
    }
}

void GameWidget::calculateBeams(void)
{
    int near_step_x = m_max_x / (num_beams - 1);
    int far_step_x = near_step_x / 3;
    int near_start_x;
    int far_start_x;
    int i;
    int j;

    if (num_beams % 2) {
	far_start_x = m_max_x / 2 - (num_beams / 2) * far_step_x;
	near_start_x = m_max_x / 2 - (num_beams / 2) * near_step_x;
    } else {
	far_start_x = m_max_x / 2 - far_step_x / 2 - ((num_beams - 1) / 2) * far_step_x;
	near_start_x = m_max_x / 2 - near_step_x / 2 - ((num_beams - 1) / 2) * near_step_x;
    }
    for (i = 0; i < num_beams; i++) {
	m_beam[i].far_x = far_start_x + i * far_step_x;
	m_beam[i].near_x = near_start_x + i * near_step_x;
    }
    m_far_y = m_max_y / 5 + 20;
    m_near_y = m_max_y;
    for (i = 0; i < num_beams; i++) {
	int k = (1000000 * (m_beam[i].near_x - m_beam[i].far_x)) / (m_near_y - m_far_y);
	int d = 1000000 * m_beam[i].far_x - k * m_far_y;
	int diff_y = (6000000 * (m_near_y - m_far_y)) / (4 * num_counter);
	int diff_diff_y = (2 * diff_y) / (3 * num_counter);
	int y = 1000000 * m_near_y;
	for (j = 0; j < num_counter; j++) {
	    m_coord[i][j].y = (y + 500000) / 1000000;
	    m_coord[i][j].x = (k * m_coord[i][j].y + d + 500000) / 1000000;
	    y = y - diff_y;
	    diff_y = diff_y - diff_diff_y;
	}
    }
}

void GameWidget::drawBeams(QPainter &painter)
{
    int far_diff_x = m_beam[1].far_x - m_beam[0].far_x;
    int near_diff_x = m_beam[1].near_x - m_beam[0].near_x;
    int far_x;
    int near_x;
    int i;

    painter.setPen(QColor(0, 0, 192));
    for (i = 0; i < num_beams; i++) {
	if (m_lefthand) {
	    painter.drawLine(m_far_y, m_max_x - m_beam[i].far_x, m_near_y, m_max_x - m_beam[i].near_x);
	} else {
	    painter.drawLine(m_max_y - m_far_y, m_beam[i].far_x, m_max_y - m_near_y, m_beam[i].near_x);
	}
    }
    far_x = m_beam[0].far_x - far_diff_x;
    near_x = m_beam[0].near_x - near_diff_x;
    do {
	if (m_lefthand) {
	    painter.drawLine(m_far_y, m_max_x - far_x, m_near_y, m_max_x - near_x);
	} else {
	    painter.drawLine(m_max_y - m_far_y, far_x, m_max_y - m_near_y, near_x);
	}
	far_x = far_x - far_diff_x;
	near_x = near_x - near_diff_x;
    } while (far_x >= 0);
    far_x = m_beam[num_beams - 1].far_x + far_diff_x;
    near_x = m_beam[num_beams - 1].near_x + near_diff_x;
    do {
	if (m_lefthand) {
	    painter.drawLine(m_far_y, m_max_x - far_x, m_near_y, m_max_x - near_x);
	} else {
	    painter.drawLine(m_max_y - m_far_y, far_x, m_max_y - m_near_y, near_x);
	}
	far_x = far_x + far_diff_x;
	near_x = near_x + near_diff_x;
    } while (far_x <= m_max_x);
    if (m_lefthand) {
	painter.drawLine(m_far_y, 0, m_far_y, m_max_x);
    } else {
	painter.drawLine(m_max_y - m_far_y, 0, m_max_y - m_far_y, m_max_x);
    }
    for (i = 0; i < 10; i++) {
	int y = getYCoord(0, i * 10 + 9 - (m_beam_counter % 10));
	if (m_lefthand) {
	    painter.drawLine(y, 0, y, m_max_x);
	} else {
	    painter.drawLine(m_max_y - y, 0, m_max_y - y, m_max_x);
	}
    }
}

void GameWidget::drawStats(QPainter &painter)
{
    const char *strings[] = {"GUNS:", "SHIELD:", "FUEL:", "BOMBS:", "LIVES:", "SCORE:"};
    int field_x = (m_max_x + 10) / 3;
    char buffer[32];
    int i;

    for (i = 0; i < 6; i++) {
	if (i % 2) {
	    drawPixmapText(painter, (i / 2) * field_x, 10, 0, strings[i]);
	} else {
	    drawPixmapText(painter, (i / 2) * field_x, 0, 0, strings[i]);
	}
	switch (i) {
	case 0:
	    drawBar(painter, (i / 2) * field_x + 50, 0, field_x - 60,
		m_gun, max_gun / 10, max_gun / 4, max_gun);
	    break;
	case 1:
	    drawBar(painter, (i / 2) * field_x + 50, 10, field_x - 60,
		m_shield, max_shield / 4, max_shield / 2, max_shield);
	    break;
	case 2:
	    drawBar(painter, (i / 2) * field_x + 50, 0, field_x - 60,
		m_fuel, max_fuel / 10, max_fuel / 4, max_fuel);
	    break;
	case 3:
	    drawBar(painter, (i / 2) * field_x + 50, 10, field_x - 60,
		m_bombs, 0, 0, max_bombs);
	    break;
	case 4:
	    drawBar(painter, (i / 2) * field_x + 50, 0, field_x - 60,
		m_lives, 1, 2, max_lives);
	    break;
	case 5:
	    snprintf(buffer, sizeof (buffer), "%d", m_score);
	    drawPixmapTextRightJustified(painter, (i / 2) * field_x + 50, 10, field_x - 60, buffer);
	    break;
	}
    }
}

void GameWidget::drawShip(QPainter &painter)
{
    int max_diff_x = (1000 * (m_player_x - m_max_x / 2) * getXCoord(0, 5)) / (m_max_x / 2);
    int diff_x;
    int i;

    if (m_state == Playing || (m_state == SafeMode && (m_counter % 4))) {
	if (max_diff_x == 0) {
	    if (m_lefthand) {
		painter.drawPixmap(m_near_y - 18, m_max_x - m_player_x - 17, m_glider_pm[1]);
	    } else {
		painter.drawPixmap(m_max_y - m_near_y, m_player_x - 17, m_glider_pm[0]);
	    }
	} else {
	    for (i = 0; i < 18; i++) {
		diff_x = (max_diff_x * i) / 17000;
		if (m_lefthand) {
		    painter.drawPixmap(m_near_y - 1 - i, m_max_x - m_player_x - 17 + diff_x,
			m_glider_pm[1], 17 - i, 0, 1, 35);
		} else {
		    painter.drawPixmap(m_max_y - m_near_y + i, m_player_x - 17 - diff_x,
			m_glider_pm[0], i, 0, 1, 35);
		}
	    }
	}
    }
}
    
void GameWidget::drawBullets(QPainter &painter)
{
    int bullet_x;
    int bullet_y;
    int size;
    int i;

    for (i = 0; i < m_bullets; i++) {
	bullet_x = getXCoord(m_bullet[i].beam, m_bullet[i].counter);
	bullet_y = getYCoord(m_bullet[i].beam, m_bullet[i].counter);
	size = (num_counter - m_bullet[i].counter) / 25;
	switch (size) {
	case 0:
	    if (m_lefthand) {
		painter.drawPixmap(bullet_y - 1, m_max_x - bullet_x - 1, m_bullet_pm[3]);
	    } else {
		painter.drawPixmap(m_max_y - bullet_y - 1, bullet_x - 1, m_bullet_pm[3]);
	    }
	    break;
	case 1:
	    if (m_lefthand) {
		painter.drawPixmap(bullet_y - 1, m_max_x - bullet_x - 1, m_bullet_pm[2]);
	    } else {
		painter.drawPixmap(m_max_y - bullet_y - 1, bullet_x - 1, m_bullet_pm[2]);
	    }
	    break;
	case 2:
	    if (m_lefthand) {
		painter.drawPixmap(bullet_y - 2, m_max_x - bullet_x - 2, m_bullet_pm[1]);
	    } else {
		painter.drawPixmap(m_max_y - bullet_y - 2, bullet_x - 2, m_bullet_pm[1]);
	    }
	    break;
	case 3:
	    if (m_lefthand) {
		painter.drawPixmap(bullet_y - 3, m_max_x - bullet_x - 3, m_bullet_pm[0]);
	    } else {
		painter.drawPixmap(m_max_y - bullet_y - 3, bullet_x - 3, m_bullet_pm[0]);
	    }
	    break;
	}
    }
}

void GameWidget::drawBomb(QPainter &painter)
{
    if (m_bomb_counter > 10) {
	if (m_lefthand) {
	    painter.drawPixmap(m_bomb_y - 5, m_max_x - m_bomb_x - 5, m_bomb_pm[1]);
	} else {
	    painter.drawPixmap(m_max_y - m_bomb_y - 5, m_bomb_x - 5, m_bomb_pm[1]);
	}
    } else if (m_bomb_counter > 0) {
	if (m_lefthand) {
	    painter.drawPixmap(m_bomb_y - 5, m_max_x - m_bomb_x - 5, m_bomb_pm[0]);
	} else {
	    painter.drawPixmap(m_max_y - m_bomb_y - 5, m_bomb_x - 5, m_bomb_pm[0]);
	}
    }
}

void GameWidget::drawMines(QPainter &painter)
{
    int mine_x;
    int mine_y;
    int size;
    int i;

    for (i = 0; i < m_mines; i++) {
	mine_x = getXCoord(m_mine[i].beam, m_mine[i].counter);
	mine_y = getYCoord(m_mine[i].beam, m_mine[i].counter);
	size = (num_counter - m_mine[i].counter) / 25;
	switch (size) {
	case 0:
	    if (m_lefthand) {
		painter.drawPixmap(mine_y - 1, m_max_x - mine_x - 1, m_mine_pm[3]);
	    } else {
		painter.drawPixmap(m_max_y - mine_y - 1, mine_x - 1, m_mine_pm[3]);
	    }
	    break;
	case 1:
	    if (m_lefthand) {
		painter.drawPixmap(mine_y - 1, m_max_x - mine_x - 1, m_mine_pm[2]);
	    } else {
		painter.drawPixmap(m_max_y - mine_y - 1, mine_x - 1, m_mine_pm[2]);
	    }
	    break;
	case 2:
	    if (m_lefthand) {
		painter.drawPixmap(mine_y - 2, m_max_x - mine_x - 2, m_mine_pm[1]);
	    } else {
		painter.drawPixmap(m_max_y - mine_y - 2, mine_x - 2, m_mine_pm[1]);
	    }
	    break;
	case 3:
	    if (m_lefthand) {
		painter.drawPixmap(mine_y - 3, m_max_x - mine_x - 3, m_mine_pm[0]);
	    } else {
		painter.drawPixmap(m_max_y - mine_y - 3, mine_x - 3, m_mine_pm[0]);
	    }
	    break;
	}
    }
}

void GameWidget::drawEnemies(QPainter &painter)
{
    int enemy_x;
    int enemy_y;
    int size;
    int i;

    for (i = 0; i < m_enemies; i++) {
	if (m_enemy[i].counter == num_counter) {
	    enemy_x = m_enemy[i].x;
	    enemy_y = m_far_y;
	} else {
	    enemy_x = getXCoord(m_enemy[i].beam, m_enemy[i].counter);
	    enemy_y = getYCoord(m_enemy[i].beam, m_enemy[i].counter);
	}
	size = (10 * (num_counter - m_enemy[i].counter)) / 125;
	switch (size) {
	case 0:
	    if (m_lefthand) {
		painter.drawPixmap(enemy_y - 2, m_max_x - enemy_x - 2, m_enemy_pm[5]);
	    } else {
		painter.drawPixmap(m_max_y - enemy_y - 2, enemy_x - 2, m_enemy_pm[5]);
	    }
	    break;
	case 1:
	    if (m_lefthand) {
		painter.drawPixmap(enemy_y - 3, m_max_x - enemy_x - 3, m_enemy_pm[4]);
	    } else {
		painter.drawPixmap(m_max_y - enemy_y - 3, enemy_x - 3, m_enemy_pm[4]);
	    }
	    break;
	case 2:
	    if (m_lefthand) {
		painter.drawPixmap(enemy_y - 4, m_max_x - enemy_x - 4, m_enemy_pm[3]);
	    } else {
		painter.drawPixmap(m_max_y - enemy_y - 4, enemy_x - 4, m_enemy_pm[3]);
	    }
	    break;
	case 3:
	    if (m_lefthand) {
		painter.drawPixmap(enemy_y - 5, m_max_x - enemy_x - 5, m_enemy_pm[2]);
	    } else {
		painter.drawPixmap(m_max_y - enemy_y - 5, enemy_x - 5, m_enemy_pm[2]);
	    }
	    break;
	case 4:
	    if (m_lefthand) {
		painter.drawPixmap(enemy_y - 6, m_max_x - enemy_x - 6, m_enemy_pm[1]);
	    } else {
		painter.drawPixmap(m_max_y - enemy_y - 6, enemy_x - 6, m_enemy_pm[1]);
	    }
	    break;
	case 5:
	    if (m_lefthand) {
		painter.drawPixmap(enemy_y - 7, m_max_x - enemy_x - 7, m_enemy_pm[0]);
	    } else {
		painter.drawPixmap(m_max_y - enemy_y - 7, enemy_x - 7, m_enemy_pm[0]);
	    }
	    break;
	}
    }
}

void GameWidget::drawBonuses(QPainter &painter)
{
    int bonus_x;
    int bonus_y;
    int size;
    int i;

    for (i = 0; i < m_bonuses; i++) {
	bonus_x = getXCoord(m_bonus[i].beam, m_bonus[i].counter);
	bonus_y = getYCoord(m_bonus[i].beam, m_bonus[i].counter);
	size = (num_counter - m_bonus[i].counter) / 25;
	switch (size) {
	case 0:
	    if (m_lefthand) {
		painter.drawPixmap(bonus_y - 2, m_max_x - bonus_x - 5, m_bonus_pm[3][m_bonus[i].what]);
	    } else {
		painter.drawPixmap(m_max_y - bonus_y - 2, bonus_x - 5, m_bonus_pm[3][m_bonus[i].what]);
	    }
	    break;
	case 1:
	    if (m_lefthand) {
		painter.drawPixmap(bonus_y - 3, m_max_x - bonus_x - 7, m_bonus_pm[2][m_bonus[i].what]);
	    } else {
		painter.drawPixmap(m_max_y - bonus_y - 3, bonus_x - 7, m_bonus_pm[2][m_bonus[i].what]);
	    }
	    break;
	case 2:
	    if (m_lefthand) {
		painter.drawPixmap(bonus_y - 4, m_max_x - bonus_x - 9, m_bonus_pm[1][m_bonus[i].what]);
	    } else {
		painter.drawPixmap(m_max_y - bonus_y - 4, bonus_x - 9, m_bonus_pm[1][m_bonus[i].what]);
	    }
	    break;
	case 3:
	    if (m_lefthand) {
		painter.drawPixmap(bonus_y - 5, m_max_x - bonus_x - 11, m_bonus_pm[0][m_bonus[i].what]);
	    } else {
		painter.drawPixmap(m_max_y - bonus_y - 5, bonus_x - 11, m_bonus_pm[0][m_bonus[i].what]);
	    }
	    break;
	}
    }
}

void GameWidget::drawExplosions(QPainter &painter)
{
    int i;

    painter.setPen(QColor(255, 255, 0));
    for (i = 0; i < m_explosions; i++) {
	if (m_explosion[i].big) {
	    if (m_lefthand) {
		painter.drawPixmap(m_explosion[i].y - 10, m_max_x - m_explosion[i].x - 20,
		    m_big_expl_pm[m_explosion[i].counter]);
	    } else {
		painter.drawPixmap(m_max_y - m_explosion[i].y - 10, m_explosion[i].x - 20,
		    m_big_expl_pm[m_explosion[i].counter]);
	    }
	} else {
	    if (m_lefthand) {
		painter.drawPixmap(m_explosion[i].y - 10, m_max_x - m_explosion[i].x - 10,
		    m_small_expl_pm[m_explosion[i].counter]);
	    } else {
		painter.drawPixmap(m_max_y - m_explosion[i].y - 10, m_explosion[i].x - 10,
		    m_small_expl_pm[m_explosion[i].counter]);
	    }
	}
    }
}

void GameWidget::drawMothership(QPainter &painter)
{
    if (m_mothership_x != -100) {
	if (m_lefthand) {
	    painter.drawPixmap(m_far_y - 7, m_max_x - m_mothership_x - 14, m_mothership_pm);
	} else {
	    painter.drawPixmap(m_max_y - m_far_y - 7, m_mothership_x - 14, m_mothership_pm);
	}
    }
}

int GameWidget::getXCoord(int beam, int counter)
{
    if (counter < num_counter) {
	return m_coord[beam][counter].x;
    } else {
	return -1;
    }
}

int GameWidget::getYCoord(int beam, int counter)
{
    if (counter < num_counter) {
	return m_coord[beam][counter].y;
    } else {
	return -1;
    }
}

void GameWidget::updateActors(void)
{
    int i;
    int j;

    /* increment = explosion counters */
    for (i = m_explosions - 1; i >= 0; i--) {
	m_explosion[i].counter++;
	if (m_explosion[i].counter >= 23) {
	    deleteExplosion(i);
	}
    }

    /* move bonuses and destroy them if they reach the screen end */
    for (i = m_bonuses - 1; i >= 0; i--) {
	m_bonus[i].counter--;
	if (m_bonus[i].counter < 10 && abs(m_beam[m_bonus[i].beam].near_x - m_player_x) < 10) {
	    switch (m_bonus[i].what) {
	    case 0:
		m_gun = max_gun;
		break;
	    case 1:
		m_shield = max_shield;
		break;
	    case 2:
		m_fuel = max_fuel;
		break;
	    case 3:
		if (m_bombs < max_bombs) {
		    m_bombs++;
		}
		break;
	    case 4:
		if (m_lives < max_lives) {
		    m_lives++;
		}
		break;
	    }
	    deleteBonus(i);
	} else if (m_bonus[i].counter == 0) {
	    deleteBonus(i);
	}
    }

    /* if the bomb counter > 0, decrement and destroy all enemies if it reaches 0 */
    if (m_bomb_counter) {
	m_bomb_counter--;
	m_bomb_x = m_bomb_x + m_bomb_diff_x;
	m_bomb_y = m_bomb_y - m_bomb_diff_y;
	if (m_bomb_counter == 0) {
	    for (i = m_enemies - 1; i >= 0; i--) {
		if (m_enemy[i].counter == num_counter) {
		    makeExplosion(m_enemy[i].x, m_far_y, 0);
		} else {
		    makeExplosion(getXCoord(m_enemy[i].beam, m_enemy[i].counter),
			getYCoord(m_enemy[i].beam, m_enemy[i].counter), 0);
		}
		deleteEnemy(i);
		m_enemies_killed++;
		addScore(50);
	    }
	    if (m_mothership_x != -100) {
		makeExplosion(m_mothership_x, m_far_y, 1);
		m_mothership_x = -100;
		m_motherships_killed++;
		addScore(500);
	    }
	}
    }

    /* every 5 seconds, remove some fuel */
    if (m_fuel > 0 && (m_counter % 100) == 0) {
	m_fuel--;
    }

    /* create and move the mothership every 60 seconds */
    if ((m_counter % 1200) == 0) {
	if (rand() % 2) {
	    m_mothership_x = m_max_x + 14;
	    m_mothership_req_x = -14;
	} else {
	    m_mothership_x = -14;
	    m_mothership_req_x = m_max_x + 14;
	}
    }
    if (m_mothership_x != -100) {
	if (m_mothership_x > m_mothership_req_x) {
	    m_mothership_x = m_mothership_x - 3;
	} else {
	    m_mothership_x = m_mothership_x + 3;
	}
	if (m_mothership_x <= -14 || m_mothership_x >= (m_max_x + 14)) {
	    m_mothership_x = -100;
	}
    }

    /* create and move the enemies */
    static int fact[] = {36, 24, 21, 18, 15, 15, 15, 12, 12, 12};
    if ((rand() % fact[(m_counter % 1200) / 120]) == 0) {
	makeEnemy();
    }
    for (i = m_enemies - 1; i >= 0; i--) {
	if (m_enemy[i].req_x != m_enemy[i].x) {
	    if (m_enemy[i].req_x > m_enemy[i].x) {
		m_enemy[i].x = m_enemy[i].x + 6;
		if (m_enemy[i].x >= (m_max_x + 7)) {
		    deleteEnemy(i);
		} else if (m_enemy[i].x > m_enemy[i].req_x) {
		    m_enemy[i].x = m_enemy[i].req_x;
		}
	    } else {
		m_enemy[i].x = m_enemy[i].x - 6;
		if (m_enemy[i].x <= -7) {
		    deleteEnemy(i);
		} else if (m_enemy[i].x < m_enemy[i].req_x) {
		    m_enemy[i].x = m_enemy[i].req_x;
		}
	    }
	} else if (m_enemy[i].req_counter != m_enemy[i].counter) {
	    if (m_enemy[i].req_counter > m_enemy[i].counter) {
		m_enemy[i].counter = m_enemy[i].counter + 2;
		if (m_enemy[i].counter > m_enemy[i].req_counter) {
		    m_enemy[i].counter = m_enemy[i].req_counter;
		}
	    } else {
		m_enemy[i].counter = m_enemy[i].counter - 3;
		if (m_enemy[i].counter < m_enemy[i].req_counter) {
		    m_enemy[i].counter = m_enemy[i].req_counter;
		}
	    }
	} else if (m_enemy[i].counter == num_counter) {
	    m_enemy[i].req_x = m_enemy[i].req_end_x;
	} else {
	    dropMine(m_enemy[i].beam, m_enemy[i].counter - 3);
	    m_enemy[i].req_counter = num_counter; 
	}
    }

    /* move the player left and right, update gun and shield strength if we have fuel */
    if (m_player_req_x > m_player_x) {
	m_player_x = m_player_x + 9;
	if (m_player_x >= m_player_req_x) {
	    m_player_x = m_player_req_x;
	    if (m_req_fire) {
		fireBullet();
	    }
	}
    } else if (m_player_req_x < m_player_x) {
	m_player_x = m_player_x - 9;
	if (m_player_x <= m_player_req_x) {
	    m_player_x = m_player_req_x;
	    if (m_req_fire) {
		fireBullet();
	    }
	}
    }
    if (m_fuel > 0 && (m_counter % 2) == 0 && m_gun < max_gun) {
	m_gun++;
    }
    if (m_fuel > 0 && (m_counter % 12) == 0 && m_shield < max_shield) {
	m_shield++;
    }

    /* move the mines until they hit the glider or the bottom or the player */
    for (i = m_mines - 1; i >= 0; i--) {
	m_mine[i].counter--;
	if (m_mine[i].counter <= 0) {
	    deleteMine(i);
	} else if (m_mine[i].counter < 10
	    && abs(m_beam[m_mine[i].beam].near_x - m_player_x) < 15) {
	    if (m_state == Playing) {
		int hit = abs(m_beam[m_mine[i].beam].near_x - m_player_x);
		if (hit < 5) {
		    m_shield = m_shield - 75;
		} else if (hit < 10) {
		    m_shield = m_shield - 50;
		} else {
		    m_shield = m_shield - 25;
		}
		if (m_shield < 0) {
		    makeExplosion(m_player_x, m_near_y - 7, 1);
		    m_player_x = m_beam[num_beams / 2].near_x;
		    m_player_req_x = m_player_x;
		    m_player_req_beam = num_beams / 2;
		    m_req_fire = 0;
		    m_bombs = 0;
		    m_lives--;
		    switchState(Dying);
		} else {
		    makeExplosion(m_player_x, m_near_y - 14, 0);
		}
	    }
	    deleteMine(i);
	}
    }

    /* move the bullets until they delete something or are too far away */
    for (i = m_bullets - 1; i >= 0; i--) {
	for (j = m_enemies - 1; j >= 0; j--) {
	    if (m_enemy[j].counter == num_counter
		&& (num_counter - m_bullet[i].counter) < 13
		&& abs(m_beam[m_bullet[i].beam].far_x - m_enemy[j].x) < 4) {
		makeExplosion(m_enemy[j].x, m_far_y, 0);
		makeBonus(m_enemy[j].beam, m_enemy[j].counter, 0);
		deleteBullet(i);
		deleteEnemy(j);
		m_enemies_killed++;
		addScore(100);
	    } else if (m_enemy[j].beam == m_bullet[i].beam
		&& abs(m_enemy[j].counter - m_bullet[i].counter) < 13) {
		if (m_enemy[j].req_counter > m_enemy[j].counter) {
		    addScore(50);
		} else {
		    addScore(100);
		}
		makeExplosion(getXCoord(m_enemy[j].beam, m_enemy[j].counter),
		    getYCoord(m_enemy[j].beam, m_enemy[j].counter), 0);
		makeBonus(m_enemy[j].beam, m_enemy[j].counter, 0);
		deleteBullet(i);
		deleteEnemy(j);
		m_enemies_killed++;
	    }
	}
	if ((num_counter - m_bullet[i].counter) < 13
	    && abs(m_beam[m_bullet[i].beam].far_x - m_mothership_x) < 8) {
	    makeExplosion(m_mothership_x, m_far_y, 1);
	    makeBonus(m_bullet[i].beam, m_bullet[i].counter, 1);
	    deleteBullet(i);
	    m_mothership_x = -100;
	    m_motherships_killed++;
	    addScore(1000);
	}
	m_bullet[i].counter = m_bullet[i].counter + 12;
	if (m_bullet[i].counter >= num_counter) {
	    deleteBullet(i);
	}
    }
}

void GameWidget::addScore(int score)
{
    int old_score = m_score;

    m_score = m_score + score;
    if (score == 1000) {
	m_lives++;
	if (m_lives > max_lives) {
	    m_lives = max_lives;
	}
	m_fuel = m_fuel + 10;
	if (m_fuel > max_fuel) {
	    m_fuel = max_fuel;
	}
    } else if (score == 500 || (m_score % 500) < (old_score % 500)) {
	m_fuel = m_fuel + 5;
	if (m_fuel > max_fuel) {
	    m_fuel = max_fuel;
	}
    } else if ((m_score % 100) < (old_score % 100)) {
	if (m_fuel < max_fuel) {
	    m_fuel++;
	}
    }
}

